import time
from string import *
from threading import BoundedSemaphore
import code
import thread
import math
import sys
from Tkinter import *

from CDPlayerGUI import CDPlayerGUI     # GUI module
from CDHardware import CDHardware	# Module for the CD hardware
from Debugger import Debugger		# Debugger

#======================================================================#
# General event handler class

class EventHandler:

 def __init__(self, filename, callback=None):
   """ filename: the description file (probably .des) to load
       callback: the callback function to be called before and after a state change
   """
   self.load_statechart(filename)
   self.state=self.find_initial_state(self.stateH)
   # The state attribute defines the current state. It is a list,
   # because there may be concurrency in the statechart.
   self.callback=callback
   self.sema=BoundedSemaphore(value=1)
   # The semaphore is to syncronize the event handler (only one event can be handled
   # at a time).
 
 def find_initial_state(self, stateH):
   """ To find the initial state(s) in the state hierachy, stateH.
   """
   states=[]
   for s in stateH.keys():
     if stateH[s]['*']:
       states.append(s)
   return states

 def is_or_is_substate(self, state1, state2):
   """ Return 1 if state1 is or is a substate of state2; 0 otherwise.
   """
   if state2=='' or (find(state1, state2)==0 and (len(state1)==len(state2) or state1[len(state2)]=='.')):
     return 1
   else:
     return 0
 
 def change_state(self, oldstate, newstate):
   """ Change the state from oldstate to newstate.
       It is the user's business to be sure the model is currently in the specified old state,
       and the state change is meaningful.
   """
   enabled=0
   delstate=[]
   for s in self.state:
     if self.is_or_is_substate(s, oldstate):
       enabled=1
       delstate.append(s)
   if enabled:
     for s in delstate:
       del self.state[index(self.state, s)]
     paths=split(newstate, '.')
     path=''
     oldpath=''
     for p in range(0, len(paths)):
       oldpath=path
       path=self.append_path(path, paths[p])
       found=0
       for p2 in self.state:	# O(n**2) of course is not a very good solution
         if self.is_or_is_substate(p2, path):
           found=1
       if not found:
         break
     state=self.stateH
     for i in range(0, p):
       state=state[paths[i]]
     self.add_states_to_path(oldpath, paths, p, state)
 
 def append_path(self, path1, path2):
   """ Append sub-path path2 to path1
   """
   if path1=='':
     return path2
   else:
     return path1+'.'+path2
 
 def add_states_to_path(self, path, paths, startpos, states):
   """ Should ONLY be called by change_state function
   """
   if startpos<len(paths):
     if states[paths[startpos]]['&']:
       for s in states.keys():
         if s!='*' and s!='&' and s!='!':
           found=0
           for s2 in self.state:
             if self.is_or_is_substate(s2, self.append_path(path, s)):
               found=1
               break
           if found:
             continue
           if s!=paths[startpos]:
             self.add_states_to_path(self.append_path(path, s), [], 0, states[s])
           else:
             self.add_states_to_path(self.append_path(path, paths[startpos]), paths, startpos+1, states[paths[startpos]])
     else:
       self.add_states_to_path(self.append_path(path, paths[startpos]), paths, startpos+1, states[paths[startpos]])
   else:
     added=0
     for s in states.keys():
       if s!='*' and s!='&' and s!='!':
         if states[s]['&'] or states[s]['*']:
           added=1
           self.add_states_to_path(self.append_path(path, s), [], 0, states[s])
     if not added:
       if not path in self.state:
         self.state.append(path)

 def add_state(self, path, line, default=0, concurrent=0, final=0):
   """ Add an empty state (which only has the 3 special attributes) to the path
   """
   state=self.stateH
   for p in path:
     state=state[p]
   newstate={'*':default}
   newstate['&']=concurrent
   newstate['!']=final
   state[line]=newstate

 def get_tabin_size(self, line):
   """ Return the tab-in size (leading spaces) of the line
   """
   for i in range(0, len(line)):
     if line[i]!=' ':
       return i
   return len(line)

 def get_space(self, num):
   """ Return a certain amount of space
   """
   s=''
   for i in range(0, num):
     s=s+' '
   return s

 def add_transition(self, event, desc):
   """ Add a transition to the list. The transition responses to the event,
       and its other attributes are specified in the dictionary desc
   """
   if desc.has_key('C'):
     cond=desc['C']
     l='vtest='
     for i in range(0, len(cond)-1):
       l=l+'('+cond[i]+') and '
     l=l+'('+cond[-1]+')'
     desc['C']=l
   if self.trans.has_key(event):
     self.InsertSortedTransList(self.trans[event], desc)
   else:
     self.trans[event]=[desc]
 
 def InsertSortedTransList(self, trans, desc):
   """ Insert desc into the transition list ordering by the depth of its old state
       trans cannot be an empty list
       This shall be done because we stipulates that if an event enables two different
       transitions at the same time, the outer one is triggered. If they are in the
       same level, the choice is random. So designers should avoid the latter case.
   """
   ddep=len(split(desc['S'], '.'))
   added=0
   for i in range(0, len(trans)):
     tdep=len(split(trans[i]['S'], '.'))
     if tdep>ddep:
       trans.insert(i, desc)
       added=1
       break
   if not added:
     trans.append(desc)

 def load_statechart(self, filename):
   """ Loads the description of a statechart from a file
   """
   f=open(filename, 'r')

   lines=f.readlines()
   for i in range(len(lines)):
     lines[i]=replace(lines[i], '\t', '        ')
     lines[i]=rstrip(replace(lines[i], '\n', ''))
     pos=find(lines[i], '#')
     if pos>=0:
       lines[i]=rstrip(lines[i][0:pos])

   self.stateH={}
   self.trans={}
   self.description=''
   parsing=None
   path=[]
   tabins={}
   lasttabin=0
   depth=0
   curevent=None

   for line in lines:
     if strip(line)=='':
       continue
     if line=='STATECHART:':
       parsing='SC'
     elif line=='TRANSITION:':
       if curevent!=None:	# There is a previous definition
         self.add_transition(curevent, definition)
       trstate=None
       parsing='TR'
       lasttabin=0
       definition={}
     elif line=='DESCRIPTION:':
       parsing='DC'
     else:
       if parsing=='SC':	# Parsing STATECHART section
         if line[len(line)-1]=='!':
           line=replace(line, '!', '')
           final=1
         else:
           final=0
         if line[len(line)-1]=='&':
           line=replace(line, '&', '')
           concurrent=1
         else:
           concurrent=0
         if line[len(line)-1]=='*':
           line=replace(line, '*', '')
           default=1
         else:
           default=0
         ti=self.get_tabin_size(line)
         line=strip(line)
         if ti in tabins.keys():
           pos=index(path, tabins[ti])
           del path[pos:]
           for k in tabins.keys():
             if k>ti:
               del tabins[k]
         else:
           for k in tabins.keys():
             if k>ti:
               print 'ERROR: Bad format in the statechart description.'
               exit
         tabins[ti]=line
         self.add_state(path, line, default, concurrent, final)
         path.append(line)
       elif parsing=='TR':	# Parsing TRANSITION section
         if len(line)<2:
           print 'ERROR: Bad format in the transition description.'
           exit
         ti=self.get_tabin_size(line)
         line=strip(line)
         f2=line[0:2]
         if f2 in ['S:', 'E:', 'C:', 'N:', 'O:']:
           # S: old State
           # E: Event
           # C: Condition (guard)
           # N: New state
           # O: Output
           lasttabin=ti+2
           trstate=f2[0]
           line=line[2:]
           ti=ti+2
         if strip(line)=='':
           continue
         else:
           if trstate==None:
             print 'ERROR: Bad format in the transition description.'
             exit
           else:
             if trstate=='E':
               curevent=strip(line)
             else:
               if ti>lasttabin:
                 l=self.get_space(ti-lasttabin)+line
               else:
                 l=line
               if trstate in ['O', 'C']:	# Output and Condition can have multiple lines
                 if definition.has_key(trstate):
                   definition[trstate].append(l)
                 else:
                   definition[trstate]=[l]
               else:
                 definition[trstate]=l
       elif parsing=='DC':			# Parsing DESCRIOTION section
         if len(self.description)>0:
           self.description=self.description+'\n'
         self.description=self.description+line

   if curevent!=None:	# There is a previous definition
     self.add_transition(curevent, definition)


 def outputFnc(self, outp):
   """ Runs the output function(s) in the list outp
   """
   ii=code.InteractiveInterpreter(sys.modules['__main__'].__dict__)
   for o in outp:
     ii.runsource(o)

 def event(self, E):
   """ Event handler
       When event E occurs, this handler is called.
       It first runs the callback function, which can be defined for debug purpose.
       After that, it decides whether a transition can be made.
       The decision is based on the current state and the definition of the transition.
       If a transition should be performed, it first runs the callback function, then changes
       the current state, carries out the output commands, and runs the callback function
       once more before return.
   """
   self.sema.acquire()
   self.callback(E, -1, '', '')
   if self.trans.has_key(E):
     desc=self.trans[E]
     handled=0
     for d in desc:
       if not (d.has_key('S') and d.has_key('N')):	# If it is not a valid transition
         continue
       for s in self.state:
         if self.is_or_is_substate(s, d['S']):
           if d.has_key('C'):		# If there is a condition defined
             ii=code.InteractiveInterpreter(sys.modules['__main__'].__dict__)
             ii.runsource(d['C'])
             if not vtest:
               continue
           # The transition can be carried out here
           self.callback(E, 1, d['S'], d['N'])
           self.change_state(d['S'], d['N'])
           if d.has_key('O'):		# If there is an output
             self.outputFnc(d['O'])
           handled=1
           self.callback(E, 0, d['S'], d['N'])
           break
       if handled:
         break
   self.sema.release()

 def ShowStateH(self):
   print '{'
   self.ShowStateH_Rec(self.stateH, 3)
   print '}'
 
 def ShowStateH_Rec(self, state, space):
   """ Shows the internal state hierachy in a more readable way (for debug purpose)
   """
   i=0
   l=len(state.keys())
   for k in state.keys():
     i=i+1
     coma=''
     if i<l:
       coma=','
     if not k in ['*', '!', '&']:
       brank=','
       if len(state[k].keys())==3:
         brank='}'+coma
       print self.get_space(space)+'\''+k+'\' : { \'*\' : '+str(state[k]['*'])+', \'!\' : '+str(state[k]['!'])+', \'&\' : '+str(state[k]['&'])+brank
       if len(state[k].keys())>3:
         self.ShowStateH_Rec(state[k], space+3)
         print self.get_space(space)+'}'+coma
         
 def ShowTransitions(self):
   """ Shows the internal transition table in a more readable way (for debug purpose)
   """
   i=0
   l=len(self.trans)
   print '{'
   for t in self.trans:
     i=i+1
     print '  \''+t+'\' : ['
     self.ShowAnEvent(self.trans[t])
     if i<l:
       print '  ],'
     else:
       print '  ]'
   print '}'
 
 def ShowAnEvent(self, trans):
   """ Should ONLY be called by ShowTransitions function
   """
   i=0
   l=len(trans)
   for t in trans:
     i=i+1
     s='    { '
     x=0
     y=len(t.keys())
     for k in t.keys():
       x=x+1
       s=s+'\''+k+'\' : '+self.ShowKey(k, t[k])
       if x<y:
         s=s+',\n      '
       else:
         s=s+' }'
     if i<l:
       s=s+','
     print s
 
 def ShowKey(self, k, desc):
   """ Should ONLY be called by ShowAnEvent function
   """
   qchar='\''
   if k in ['S', 'N', 'C']:
     if find(desc, qchar)>=0:
       qchar='"'
     return qchar+desc+qchar
   else:
     s='[ '
     i=0
     j=len(desc)
     for d in desc:
       if find(d, qchar)>=0:
         qchar='"'
       i=i+1
       s=s+qchar+d+qchar
       if i<j:
         s=s+',\n              '
       qchar='\''
     s=s+' ]'
     return s

#======================================================================#
# Clock class

class Clock:
 """ The Clock is a scheduler running in another thread. When it is time
     to raise an event, it runs the function func with parameter param.
 """

 def start(self, inteval, func, param):
   self.stopped=0
   thread.start_new_thread(self.ClockThread, (inteval, func, param))
 
 def stop(self):
   self.stopped=1
 
 def ClockThread(self, inteval, func, param):
   while not self.stopped:
     time.sleep(inteval)
     if not self.stopped:
       func(param)

#======================================================================#
# Simulator

debugger=Debugger()

eventhandler=EventHandler('CDPlayer.des', debugger.EventDebugger)
debugger.SetEventHandler(eventhandler)
print eventhandler.description

# When the display is blanking, we want to trace the change of states
debugger.TraceEvent('BlankTime', 0, 'Started.CDIn.Paused')
# When the debug hotkey is pressed, the program is interrupted and the debug mode is entered
debugger.InterruptEvent('GUI Debug', -1)

cd = CDHardware(eventhandler, 0)

root = Tk()
root.title("CD player")
gui = CDPlayerGUI(root, eventhandler)

eventhandler.event('Start')

root.mainloop()