import code
import sys
from threading import BoundedSemaphore
import time
import thread

from string import *

class Debugger:

  def __init__(self):
    self.trace_list=[]
    self.interrupt_list=[]
    self.eventhandler=None

  def SetEventHandler(self, eventhandler):
    self.eventhandler=eventhandler

  def TraceEvent(self, event, before=0, oldstate='', newstate=''):
    self.trace_list.append({'E':event, 'B':before, 'O':oldstate, 'N':newstate})

  def InterruptEvent(self, event, before=1, oldstate='', newstate=''):
    self.interrupt_list.append({'E':event, 'B':before, 'O':oldstate, 'N':newstate})

  def CheckEvent(self, event, before, oldstate, newstate, elist):
    """ before: a flag indicating the point where the event handler is in.
                -1: the event handler is starting; it still doesn't know whether the transition
                    will take place or not (so the oldstate and newstate had no meaning);
                 0: the event handler decided to make a transition; the transition is about to
                    take place;
                 1: the event handler decided to make a transition; the transition has been
                    complished.
    """
    if self.eventhandler!=None:
      for i in elist:
        if i['E']==event and i['B']==before and self.eventhandler.is_or_is_substate(oldstate, i['O']) and self.eventhandler.is_or_is_substate(newstate, i['N']):
          return 1
    return 0
 
  def EventDebugger(self, event, before, oldstate, newstate):
    if self.CheckEvent(event, before, oldstate, newstate, self.trace_list):
      print 'Debugger: (Trace)'
      print '  Event \''+event+'\' occurred.'
      print '  Old state: '+str(oldstate)
      print '  New state: '+str(newstate)
      print '  Current state of the event handler:\n    '+str(self.eventhandler.state)
    if self.CheckEvent(event, before, oldstate, newstate, self.interrupt_list):
      print 'Debugger: (Interrupt)'
      print '  Event \''+event+'\' occurred.'
      print 'Interrupt mode: enter command \'exit\' to leave.'
      ii=code.InteractiveInterpreter(sys.modules['__main__'].__dict__)
      while 1:
        sys.__stdout__.write('>>> ')
        prompt=strip(sys.__stdin__.readline())
        if prompt=='exit':
          break;
        err=ii.runsource(prompt)
        if err==1:
          prompt2='.'
          while prompt2!='':
            sys.__stdout__.write('    ')
            prompt2=rstrip(sys.__stdin__.readline())
            if strip(prompt2)=='':
              break;
            prompt=prompt+'\n'+prompt2
          prompt=prompt+'\n'
          err=ii.runsource(prompt)
          
      print 'On leaving interrupt mode, the program is re-activated.'